package com.albertoventurini.parsley.examples.graphml;

import com.albertoventurini.parsley.grammar.Grammar;
import com.albertoventurini.parsley.grammar.ParseTree;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class GrammarGraphMLParser implements GraphMLParser {

    private final Grammar grammar = new GraphMLGrammar();

    @Override
    public GraphMLParseResult parse(final String filePath) {
        try {
             return grammar
                     .parse(Files.readString(Path.of(filePath)))
                     .map(this::processParseTree)
                     .orElseThrow(() -> new RuntimeException("Unable to parse graphml file: " + filePath));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private GraphMLParseResult processParseTree(final ParseTree parseTree) {

        final List<Node> nodes = parseTree.getFirstDescendantByName("nodes").map(n ->
            n.getChildren()
                    .stream()
                    .map(this::processNodeElem)
                    .collect(Collectors.toList())
        ).orElseThrow();

        final List<Edge> edges = parseTree.getFirstDescendantByName("edges").map(n ->
                n.getChildren()
                        .stream()
                        .map(this::processEdgeElem)
                        .collect(Collectors.toList())
        ).orElseThrow();

        return new GraphMLParseResult(nodes, edges);
    }

    private Node processNodeElem(final ParseTree nodeElem) {
        final String id = extractAttribute(nodeElem.child("tag:node"), "id");

        final Map<String, String> data = nodeElem.child("content").getChildren()
                .stream()
                .map(this::extractDataElemContent)
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));

        return new Node(id, data);
    }

    private Edge processEdgeElem(final ParseTree edgeElem) {
        final String id = extractAttribute(edgeElem.child("tag:edge"), "id");
        final String source = extractAttribute(edgeElem.child("tag:edge"), "source");
        final String target = extractAttribute(edgeElem.child("tag:edge"), "target");

        final Map<String, String> data = edgeElem.child("content").getChildren()
                .stream()
                .map(this::extractDataElemContent)
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));

        return new Edge(id, source, target, data);
    }

    private Map.Entry<String, String> extractDataElemContent(final ParseTree dataElem) {
        final var attribute = extractAttribute(dataElem.child("tag:data"), "key");
        final var content = dataElem.child(1).getText();
        return Map.entry(attribute, content);
    }

    private String extractAttribute(final ParseTree tag, final String attributeKey) {
        return tag.child("attributes").getChildren()
                .stream()
                .filter(a -> a.child(0).getText().equals(attributeKey))
                .map(a -> a.child(1).getText())
                .findFirst()
                .orElseThrow();
    }
}
